#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 21:07:55 2020

@author: carl
"""

import os
import sys
import traceback
import multiprocessing
import signal

from PyQt5.QtWidgets import QMainWindow, QApplication, QFileDialog
from PyQt5.QtWidgets import QErrorMessage, QMessageBox, QStyleFactory
from PyQt5.QtCore import Qt, QSettings, QLocale
from PyQt5.QtGui import QFontDatabase, QFont, QFontInfo
from PyQt5.Qt import qApp

from .exceptiondialog import ExceptionDialog
from .copyfigure import add_clipboard_to_figures
from .. import octavvs_version



class OctavvsMainWindow(QMainWindow):

    fileOptions = QFileDialog.Options() | QFileDialog.DontUseNativeDialog

    def __init__(self, parent=None):
        super().__init__(parent)
        self.settings = QSettings('MICCS', 'OCTAVVS ' + self.program_name())
        qApp.installEventFilter(self)
        self.setupUi(self)

        self.errorMsg = QErrorMessage(self)
        self.errorMsg.setWindowModality(Qt.WindowModal)

        self.units = 'cm-1'
        self.plot_ltr = True

    def post_setup(self):
        "Called after setting up UI but before loading data etc"
        self.setWindowTitle('OCTAVVS %s %s' % (self.program_name(), octavvs_version))
        ExceptionDialog.install(self)

    @classmethod
    def program_name(cls):
        "Return the name of the program that this main window represents"
        return cls.__name__

    def showDetailedErrorMessage(self, err, details):
        "Show an error dialog with details"
        q = QMessageBox(self)
        q.setIcon(QMessageBox.Critical)
        q.setWindowTitle("Error")
        q.setText(err)
        q.setTextFormat(Qt.PlainText)
        q.setDetailedText(details)
        q.addButton('OK', QMessageBox.AcceptRole)
        return q.exec()

    def loadErrorBox(self, file, err, warning=False):
        "Prepare an error dialog for failure to load a file"
        q = QMessageBox(self)
        if warning:
            q.setIcon(QMessageBox.Information)
            q.setWindowTitle("Warning loading file")
            q.setText("Warning from loading '"+file+"':\n"+err[0])
        else:
            q.setIcon(QMessageBox.Warning)
            q.setWindowTitle("Error loading file")
            q.setText("Failed to load '"+file+"':\n"+err[0])
        q.setTextFormat(Qt.PlainText)
        q.setDetailedText(err[1])
        return q

    def getLoadSaveFileName(self, title, filter=None, settingname=None,
                            savesuffix=None, multiple=False,
                            settingdefault=None,
                            directory=None, defaultfilename=None):
        "Show a file dialog and select one or more files"
        setting = self.settings.value(settingname, settingdefault
                                      ) if settingname is not None else None
        if directory is None:
            directory = setting if type(setting) is str else None
        dialog = QFileDialog(parent=self, caption=title,
                             directory=directory, filter=filter)
        if defaultfilename is not None:
            dialog.selectFile(defaultfilename)
#        if setting and type(setting) is not str:
#            dialog.restoreState(setting)
        dialog.setOption(QFileDialog.DontUseNativeDialog, True)
        if savesuffix is not None:
            dialog.setAcceptMode(QFileDialog.AcceptSave)
            dialog.setDefaultSuffix(savesuffix)
        elif multiple:
            dialog.setFileMode(QFileDialog.ExistingFiles)
        else:
            dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.exec()
        files = dialog.selectedFiles()
        if not dialog.result() or not files:
            return None
        if settingname is not None:
            self.settings.setValue(settingname, os.path.dirname(files[0]))
        return files if multiple else files[0]

    def getSaveFileName(self, title, **kwargs):
        "Show a file dialog and select an output file"
        return self.getLoadSaveFileName(title=title, **kwargs)

    def getLoadFileName(self, title, **kwargs):
        "Show a file dialog and select an input file"
        return self.getLoadSaveFileName(title=title, **kwargs)

    def getLoadFileNames(self, title, **kwargs):
        "Show a file dialog and select an input file"
        return self.getLoadSaveFileName(title=title, multiple=True, **kwargs)

    def getImageFileName(self, title, **kwargs):
        "Show a file dialog and select a single image file"
        return self.getLoadSaveFileName(title=title,
                filter="Image files (*.jpg *.jpeg *.png *.tif "
                    "*.tiff *.bmp *.gif);;All files (*)",
                **kwargs)

    def getDirectoryName(self, title, settingname=None, savesetting=True,
                         default=None):
        "Show a file dialog and select a directory"
        directory = self.settings.value(settingname, default) if \
            settingname is not None else None
        dialog = QFileDialog(parent=self, caption=title,
                             directory=directory)

        dialog.setAcceptMode(QFileDialog.AcceptSave)
        dialog.setFileMode(QFileDialog.Directory)
        dialog.setOption(QFileDialog.DontUseNativeDialog, True)
        dialog.setOption(QFileDialog.ShowDirsOnly, True)
        dialog.exec()
        dirs = dialog.selectedFiles()
        if not dialog.result() or not dirs:
            return None
        if savesetting:
            self.settings.setValue(settingname, dirs[0])
        return dirs[0]

    def updateWavenumberRange(self):
        "For derived classes, when the wn range is updated"
        pass

    def updateDimensions(self, wh):
        "For derived classes, when image width/height is updated"
        pass

    def updateUnitsAndOrder(self, units, ltr):
        "Update the units (cm^-1 or nm) and direction of plots"
        self.units = units
        self.plot_ltr = ltr

    def unitsRichText(self):
        if self.units == 'cm-1':
            return 'cm<sup>-1</sup>'
        return self.units


    @staticmethod
    def choose_font(families):
        db = QFontDatabase()
        for family in families:
            if family in db.families():
                font = QFont(family)
                font.setStyleHint(QFont.SansSerif, QFont.PreferDefault)
                font.setStyleStrategy(
                    QFont.PreferAntialias | QFont.NoFontMerging)
                return font
        return QFont()  # system default

    @classmethod
    def run_octavvs_application(windowclass, parser=None, parameters=[],
                                isChild=False):
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        res = 1
        try:
            progver = f"OCTAVVS {windowclass.program_name()} {octavvs_version}"
            windowparams = {}
            fontsize = 11
            native_style = True
            if parser is not None:
                parser.add_argument('--version', action='version',
                                    version=progver)
                parser.add_argument('--font-size', metavar="pts",
                                    help="font size in points",
                                    default=fontsize,
                                    type=int)
                parser.add_argument('--native-style',
                                    help="don't switch to Fusion QStyle",
                                    action='store_true')
                selmp = multiprocessing.get_start_method(
                    allow_none=True) is None
                if selmp:
                    parser.add_argument(
                        '--mpmethod', help="method used to start child"
                        " processes: 'fork', 'spawn' or 'forkserver'")
                args = parser.parse_args()
                fontsize = args.font_size
                native_style = args.native_style
                windowparams = { k: args.__dict__[k] for k in parameters }
                if selmp and args.mpmethod:
                    multiprocessing.set_start_method(args.mpmethod)

            app = QApplication.instance()
            if not app:
                app = QApplication(sys.argv)
                if not native_style:
                    QLocale.setDefault(QLocale.c())
                    app.setStyle(QStyleFactory.create("Fusion"))
                font = OctavvsMainWindow.choose_font(
                    ["Segoe UI", "Noto Sans", "Arial"])
                font.setPointSize(fontsize)
                app.setFont(font)
                i = QFontInfo(font)
                print(f"OCTAVVS GUI font set to {i.family()}, {i.pointSize()}"
                      f" ({i.pixelSize()} px)")

            add_clipboard_to_figures()
            window = windowclass()
            window.show()
            window.post_setup(**windowparams)
            if not isChild:
                qApp.lastWindowClosed.connect(qApp.quit);
                res = app.exec_()

        except Exception:
            traceback.print_exc()
            print('Press some key to quit')
            input()
        if not isChild :
            sys.exit(res)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 18 14:45:53 2021

@author: carl
"""

import numpy as np
import scipy, scipy.optimize, scipy.spatial.distance
import math
import time
import os
from threadpoolctl import threadpool_limits
from functools import partial
from .util import pca_nipals

try:
    # os.environ["NUMBA_DISABLE_INTEL_SVML"] = "1"
    import numba
    import numba_nnls
    nnls = numba_nnls.nnls_007_111
    # nnls = numba_nnls.nnls_112_114 # crashes? icc_rt?
    use_numba = True
except ModuleNotFoundError:
    nnls = scipy.optimize.nnls
    use_numba = False
    print("INFO: numba_nnls not found; for improved MCR-ALS performance "
          "consider installing https://github.com/Nin17/numba-nnls")

def simplisma(d, nr, f):
    """
    The SIMPLISMA algorithm for finding a set of 'pure' spectra to serve
    as starting point for MCR-ALS etc.
    Reference Matlab Code:
        J. Jaumot, R. Gargallo, A. de Juan, R. Tauler,
        Chemometrics and Intelligent Laboratoty Systems, 76 (2005) 101-110

    Parameters
    ----------
    d : array(nwavenums, nspectra)
        input spectra.
    nr : int
        number of output components.
    f : float
        noise threshold.

    Returns
    -------
    spout: array(nr, nspectra)
        concentration profiles of 'purest' spectra.
    imp : array(nr, dtype=int)
        indexes of the 'purest' spectra.
    """

    nrow = d.shape[0]
    ncol = d.shape[1]
    s = d.std(axis=0)
    m = d.mean(axis=0)
    mf = m + m.max() * f
    p = s / mf

    # First Pure Spectral/Concentration profile
    imp = np.empty(nr, dtype=np.int64)
    imp[0] = p.argmax()

    #Calculation of correlation matrix
    l2 = s**2 + mf**2
    dl = d / np.sqrt(l2)
    c = (dl.T @ dl) / nrow

    #calculation of the first weight
    w = (s**2 + m**2) / l2
    p *= w
    #calculation of following weights
    dm = np.zeros((nr+1, nr+1))
    for i in range(1, nr):
        dm[1:i+1, 1:i+1] = c[imp[:i], :][:, imp[:i]]
        for j in range(ncol):
            dm[0, 0] = c[j, j]
            dm[0, 1:i+1] = c[j, imp[:i]]
            dm[1:i+1, 0] = c[imp[:i], j]
            w[j] = np.linalg.det(dm[0:i+1, 0:i+1])
        imp[i] = (p * w).argmax()

    ss = d[:,imp]
    spout = ss / np.sqrt(np.sum(ss**2, axis=0))
    return spout.T, imp

def simplisma_with_reference(data, nr, f, refdata):
    """
    Like @simplisma but with m reference spectra that become the first
    m spectra of the output.

    Parameters
    ----------
    data : array(nspectra, nwavenums)
        Input spectra. Note the transposed order vs simplisma.
    refdata : array(nrefspectra, nwavenums)
        Reference spectra.
    nr : int
        Number of output components, nr >= len(refdata)
    f : float
        Noise threshold.

    Returns
    -------
    spout: array(nr, nspectra)
        concentration profiles of reference and 'purest' spectra.
    """
    s_comps = nr - len(refdata)
    if s_comps < 0:
        raise ValueError("Number of simplisma output spectra cannot be "
                         "lower than number of reference spectra")
    if s_comps == 0:
        return refdata.copy()

    # s_data = data
    # for i in range(len(refdata)-1, 0-1, -1):
    #     proj = np.dot(s_data, refdata[i]) / (refdata[i] ** 2).sum()**.5
    #     print("proj before",i, (proj**2).sum()**.5)
    #     s_data = s_data - proj[:, None] @ refdata[i][None, :]
    # for i in range(len(refdata)-1, 0-1, -1):
    #     proj = np.dot(s_data, refdata[i]) / (refdata[i] ** 2).sum()**.5
    #     print("proj after",i, (proj**2).sum()**.5)
    # s_init = simplisma(s_data.T, s_comps, f)
    # return np.vstack((refdata, data[s_init[1]]))

    s_data = data
    for i in range(len(refdata)-1, 0-1, -1):
        proj = np.dot(s_data, refdata[i]) / (refdata[i] ** 2).sum()**.5
        s_data = s_data - proj[:, None] @ refdata[i][None, :]

    # Components first, then spectra
    c_other = simplisma(s_data, s_comps, f)[0]
    # s_init = np.linalg.lstsq(s_pre.T, data, rcond=-1)[0]
    # import matplotlib.pyplot as plt
    # plt.figure("initcomps")
    # plt.cla()
    # plt.plot(c_other.T)

    s_init = []
    for i in range(s_data.shape[1]):
        r, _ = nnls(c_other.T, s_data[:, i])
        s_init.append(r)
    s_init = np.array(s_init).T
    s_init = s_init / s_init.mean(1, keepdims=True)
    return np.vstack((refdata, s_init))

    # c_ref = []
    # for i in range(len(data)):
    #     r, _ = scipy.optimize.nnls(refdata.T, data[i])
    #     c_ref.append(r)
    # print("cs", np.array(c_ref).T.shape, c_other.shape)
    # comps = np.vstack((np.array(c_ref).T, c_other))
    # import matplotlib.pyplot as plt
    # plt.figure("initcomps")
    # plt.cla()
    # plt.plot(comps.T)
    # s_init = []
    # for i in range(data.shape[1]):
    #     r, _ = scipy.optimize.nnls(comps.T, data[:, i])
    #     s_init.append(r)
    # print("s", np.array(s_init).T.shape, refdata.shape)
    # s_init = np.array(s_init).T
    # s_init[:len(refdata), :] = refdata
    # s_init = s_init / s_init.mean(1, keepdims=True)

    return s_init


def clustersubtract(data, components, skewness=100, power=2, verbose=False):
    """
    Create initial spectra for MCR-ALS based on successively removing
    what appears to be the strongest remaining component.

    Parameters
    ----------
    data : array (nspectra, nfeatures)
        Spectral data.
    components : int
        Number of components to return.
    skewness : float, optional
        Asymmetry between positive and negative residuals when computing
        how much of the previous component to remove from the data.
        The default is 100.
    power : float, optional
        The sum of residuals is raised to this power before summation to
        determine the leading remaining component.

    Returns
    -------
    initial_spectra : array (components, nfeatures)
    """
    def typical_cluster(data, first):
        # draw sqrt(n) random data points, r
        # find closest in r for each s in data
        # for r with most s, return mean of those s
        r = np.random.choice(
            len(data), min(200, math.floor(math.sqrt(len(data)))))
        rd = data[r]
        nearest = scipy.spatial.distance.cdist(
            rd, data, 'cosine').argmin(axis=0)
        # Mean of those who are nearest the biggest cluster
        if first:
            selected = np.bincount(nearest).argmax()
        else:
            sums = data.sum(1) ** power
            selected = np.bincount(nearest, weights=sums).argmax()
        return data[nearest == selected].mean(0)

    comps = []
    sgn = np.ones_like(data, dtype=bool)
    for c in range(components):
        tc = typical_cluster(data, c == 0)
        tc = np.maximum(tc, 0)
        tc = tc / (tc * tc).sum() ** .5

        comps.append(tc)
        for i in range(10):
            ww = 1 * sgn + skewness * ~sgn
            a = (data * ww * tc).sum(1) / (ww * tc * tc).sum(1)
            oldsgn = sgn
            sgn = data > a[:, None] @ tc[None, :]
            if verbose:
                chg = (sgn != oldsgn).sum()
                print(f'clsub iter {c:3d} {i:2d} changed {chg}')
            if np.array_equal(sgn, oldsgn):
                break
        data = data - a[:, None] @ tc[None, :]
    return np.array(comps)

def numpy_scipy_threading_fix_(func):
    """
    This decorator for mcr_als prevents threading in BLAS if scipy's old NNLS
    is used, because for some reason NNLS won't be parallelized if called
    shortly after lstsq or @. This makes a *massive* difference to the
    time needed for Anderson acceleration, where the BLAS calls themselves
    take negligible time. For mixed NNLS/lstsq solving (of MCR-ALS on
    derivatives) it's less obvious whether NNSL or lstsq should be allowed
    to be parallelized.
    Note: This issue is seen on my Linux setups and might be highly version-
    dependent.
    Update 2025: The rewritten non-Fortran NNLS doesn't have this problem
    but is slower overall. Solution: Numba and numba_nnls,
    https://github.com/Nin17/numba-nnls
    """
    def check(*args, **kwargs):
        th = kwargs.get('blas_threads', 1)
        kwargs.pop('blas_threads', None)
        if (('nonnegative' not in kwargs or np.any(kwargs['nonnegative']))
            and th):
            with threadpool_limits(th, 'blas'):
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return check

if use_numba:
    @numba.jit
    def multi_nnls(A : np.ndarray, bs : np.ndarray, out : np.ndarray):
        err2 = 0
        for i in range(len(bs)):
            out[i], res = nnls(A, bs[i])
            err2 = err2 + res * res
        return err2
else:
    def multi_nnls(A : np.ndarray, bs : np.ndarray, out : np.ndarray):
        """
        Run NNLS on multiple vectors.

        Parameters
        ----------
        A : np.ndarray
            (m, n) input to nnls.
        bs : np.ndarray
            (k, m) set of k vectors.
        out : np.ndarray
            (k, n) set of k solution vectors.
        Returns
        -------
        err2 : float
            Sum of all squared residuals.
        """
        err2 = 0
        for i in range(len(bs)):
            out[i], res = nnls(A, bs[i])
            err2 = err2 + res * res
        return err2


# @numpy_scipy_threading_fix_
def mcr_als(sp, initial_A, *, maxiters=100, nonnegative=(True, True),
            tol_abs_error=0, tol_rel_improv=None, tol_rel_iters=10,
            tol_iters_after_best=None,
            maxtime=None, callback=None, acceleration=None, normalize=None,
            fixed_components_A=None, fixed_features_A=None,
            mutable_components_B=slice(None),
            initial_B=None,
            modify_before_Astep=None,
            contrast_weight=None, return_time=False,
            starttime=None):
    """
    Perform MCR-ALS nonnegative matrix decomposition on the matrix sp.
    The input array is decomposed as A@B where A and B are spectra and
    concentrations or vice versa.

    Parameters
    ----------
    sp : array(nsamples, nfeatures)
        Spectra to be decomposed.
    initial_A : array(ncomponents, nfeatures)
        Initial spectra (or concentrations).
    maxiters : int
        Maximum number of iterations.
    nonnegative : pair of bool, default (True, True)
        True if (initial, other) components must be non-negative
    tol_abs_error : float, optional
        Error target (mean square error).
    tol_rel_improv : float, optional
        Stop at average relative improvement of less than tol_rel_improv.
    tol_rel_iters : int, optional
        Number of iterations to evaluate tol_rel_improv over.
    tol_iters_after_best : int, optional
        Stop after this many iterations since last best error.
    maxtime : float, optional
        Stop after this many seconds of process time have elapsed
    callback : func(it : int, err : list(float), A : array, B : array)
        Progress callback for every iteration.
    acceleration : str, optional
        None or 'Anderson'.
        Anderson acceleration operates on whole iterations (A or B updates),
        mixing earlier directions to step towards the fixed point. This
        implementation restarts from basic updates when those would be
        better.
    normalize : str, optional
        Which matrix to l2 normalize: None, 'A' or 'B'
    fixed_components_A : int, optional
        Number of components of initial_A that are kept as-is; e.g.
        n endmembers for unmixing. The first n components are kept unchanged
        and will be present in the returned A and B. The caller will have to
        reorder the input/output as needed.
    fixed_features_A : int, optional
        Number of initial features of sp and A that are kept as-is. These
        features will typically be used to push B in a desired direction.
    modify_before_Astep : None or func(A : array, B : array, sp : array)
        Function called before updates of A, intended for rewriting of the
        (original) input data (sp) or the B array, e.g. for smoothing of
        B or updating averaged target values.
    mutable_components_B : slice, range or bool array, optional
        Components of B that will be updated. Default: all components.
        This is useful when some components have already been fitted.
    initial_B : array(ncomponents, nsamples), optional
        Necessary when mutable_components_B is used; the values for the
        non-mutable components will be taken from here.
    contrast_weight : (str, float), optional, deprecated
        Increase contrast in one matrix by mixing the other, named matrix
        ('A' or 'B') with the mean of its vectors. If A is spectra,
        try contrast_weight=('B', 0.05) to increase spectral contrast.
        See Windig and Keenan, Applied Spectroscopy 65: 349 (2011).
    return_time : bool, default False
        Measure and return process_time at each iteration.

    Returns
    -------
    A : array(ncomponents, nfeatures)
        Spectra (at lowest error)
    B : array(ncomponents, nsamples)
        Concentrations at lowest error
    error : list(float)
        Mean square error at every iteration
    process_time : list(float)
        Time relative start at each iteration, only if return_time is True.
    """
    if normalize not in [None, 'A', 'B']:
        raise ValueError("Normalization must be None, 'A' or 'B'")

    nsamp, nfeat = sp.shape
    ncomp = initial_A.shape[0]

    if initial_A.shape[1] != nfeat:
        raise ValueError("Mismatching size of data and initial vectors; "
                         f"{sp.shape} vs {initial_A.shape}")
    iAT = initial_A.T
    nfixed_a = 0
    if fixed_components_A is not None:
        nfixed_a = fixed_components_A
    if nfixed_a > ncomp:
        raise ValueError(
            f"The number of fixed components of A ({nfixed_a}) "
            f"cannot exceed the total number of components ({ncomp})")
    nfreeze = 0
    if fixed_features_A is not None:
        nfreeze = fixed_features_A
    if normalize == 'A':
        A = np.zeros(iAT.shape)
        A[:nfreeze] = iAT[:nfreeze]
        norm = np.linalg.norm(iAT[nfreeze:], axis=0)
        np.divide(iAT[nfreeze:], norm, where=norm!=0, out=A[nfreeze:])
    else:
        A = iAT.copy()

    B = np.empty((ncomp, nsamp)) if initial_B is None else initial_B.copy()
    if mutable_components_B is not None:
        fixed_B = np.ones(ncomp, dtype=bool)
        fixed_B[mutable_components_B] = False
    else:
        fixed_B is None

    newA = newB = None
    errors = []
    errorbest = None
    error = preverror = None

    cw = 0
    if contrast_weight is not None:
        if contrast_weight[0] == 'A':
            cw = contrast_weight[1]
        elif contrast_weight[0] == 'B':
            cw = -contrast_weight[1]
        else:
            raise ValueError("contrast_weight must be ('A'|'B', [0-1])")

    if acceleration == 'Anderson':
        ason_m = 2 # Memory depth
        ason_Bmode = False # Currently accelerating B steps?
        ason_g = None
        ason_G = []
        ason_X = []
    elif acceleration:
        raise ValueError("acceleration must be None or 'Anderson'")

    if starttime is None:
        starttime = time.process_time()
    times = []

    # Loop logic:
    # Half-iterations with ba=0 (update A) and ba=1 (update B).
    # Take a half-step (A or B):
    #   Update B from (prevA := newA (if set) or A) [or vice versa A<->B]
    #   Forget newA
    # If no acceleration: just keep iterating.
    # If currently accelerating the just-updated A/B:
    #   If error got worse:
    #       Reset acceleration memory
    #       redo the step from A/B (not newA/B)
    #       switch to acceleration mode A<->B [can be optional]
    #   else if not (very first iteration and accelerating B):
    #       First acc. step:
    #           Just save delta-A (X) [or B]
    #       Else:
    #           save delta-delta (G), then compute accelerated newA [or newB]
    # So: A is ok. newA is proposed accelerated A. Evaluated to make B which
    #   is rejected if error grows compared with A; then a normal step from
    #   A to B is guaranteed to be ok. Acceleration continues from that B.
    for it in range(maxiters):
        ba = 0
        retry = False
        while ba < 2:
            if not retry:# unnecessary?
                preverror = error
            if ba == 0:
                if newA is None:
                    newA = A
                prevA = newA
                if cw > 0:
                    newA = (1 - cw) * newA + cw * newA.mean(1)[:,None]
                if fixed_B is not None:
                    tmpsp = sp - (newA[:, fixed_B] @ B[fixed_B]).T
                else:
                    tmpsp = sp
                if nonnegative[1]:
                    if fixed_B is None:
                        tmpB = np.empty_like(B.T)
                        error = multi_nnls(newA, tmpsp, tmpB)
                        B = tmpB.T
                    else:
                        if not retry: # only needed when ason_Bmode==0 ?
                            B = np.copy(B)
                        tmpB = np.empty(shape=B[mutable_components_B].T.shape)
                        error = multi_nnls(newA[:, mutable_components_B],
                                           tmpsp, tmpB)
                        B[mutable_components_B] = tmpB.T
                else:
                    if not retry: # only needed when ason_Bmode==0 ?
                        B = np.copy(B)
                    B[mutable_components_B], res, _, _ = np.linalg.lstsq(
                        newA[:, mutable_components_B], tmpsp, rcond=-1)
                    error = res.sum()
                if normalize == 'B':
                    norm = np.linalg.norm(
                        B[mutable_components_B], axis=1, keepdims=True)
                    np.divide(B[mutable_components_B], norm, where=norm!=0,
                              out=B[mutable_components_B])

                newA = None
            else:
                if newB is None:
                    newB = B
                prevB = newB
                if cw < 0:
                    newB = (1 + cw) * newB - cw * newB.mean(1)[:,None]
                if modify_before_Astep is not None:
                    modify_before_Astep(A, newB, sp)

                if nfixed_a == ncomp:
                    # Single pass, nothing more to fit.
                    error = ((sp[:, nfreeze:] - (A[nfreeze:] @ newB).T
                             )**2).sum()
                    break
                elif nfixed_a > 0:
                    tmpsp = sp[:, nfreeze:] - (
                        A[nfreeze:, :nfixed_a] @ newB[:nfixed_a]).T
                else:
                    tmpsp = sp[:, nfreeze:]

                if nonnegative[0]:
                    if not retry: # only when ason_Bmode == 1 ?
                        A = A.copy()
                    error = multi_nnls(newB[nfixed_a:, :].T,
                                       tmpsp.T, A[nfreeze:, nfixed_a:])
                else:
                    tmpA, res, _, _ = np.linalg.lstsq(
                        newB[nfixed_a:, :].T, tmpsp, rcond=-1)
                    A[nfreeze:, nfixed_a:] = tmpA.T
                    error = res.sum()

                if normalize == 'A':
                    norm = np.linalg.norm(A[:, nfixed_a:], axis=0)
                    np.divide(A[:, nfixed_a:], norm, where=norm!=0,
                              out=A[:, nfixed_a:])
                newB = None

            if acceleration is None:
                pass
            elif ba == ason_Bmode:
                if retry:
                    retry = False
                    ason_Bmode = not ason_Bmode # Alternate A/B acceleration
                elif len(ason_X) > 1 and error > preverror:
                    ason_X = []
                    ason_G = []
                    retry = True
                    ba = ba - 1
                else:
                    pass
            elif ason_Bmode == 1 and it == 0:
                pass
            else:
                prevg = ason_g
                ason_g = ((A - prevA) if ba else (B - prevB)).flatten()
                if len(ason_X) < 1:
                    ason_X.append(ason_g)
                else:
                    ason_G.append(ason_g - prevg)
                    while(len(ason_G) > ason_m):
                        ason_G.pop(0)
                        ason_X.pop(0)
                    Garr = np.asarray(ason_G)
                    try:
                        gamma = np.linalg.lstsq(Garr.T, ason_g, rcond=None)[0]
                    except np.linalg.LinAlgError:
                    #     gamma = scipy.linalg.lstsq(Garr.T, ason_g)[0]
                    # except scipy.linalg.LinAlgError:
                        print('lstsq failed to converge; '
                              'restart at iter %d' % it)
                        ason_X = []
                        ason_G = []
                    else:
                        dx = ason_g - gamma @ (np.asarray(ason_X) + Garr)
                        ason_X.append(dx)
                        if ba:
                            newA = prevA + dx.reshape(A.shape)
                            if nonnegative[0]:
                                np.maximum(0, newA, out=newA)
                        else:
                            newB = prevB + dx.reshape(B.shape)
                            if nonnegative[1]:
                                np.maximum(0, newB, out=newB)
            ba = ba + 1

        curtime = time.process_time() - starttime
        times.append(curtime)
        errors.append(error)
        if not it or error < errorbest:
            errorbest = error
            Abest = A
            Bbest = B
            iterbest = it
            if nfixed_a == ncomp:
                break
        if it:
            if error < tol_abs_error:
                break
            if tol_rel_improv and it > tol_rel_iters:
                emax = max(errors[-tol_rel_iters-1:-2])
                if (emax - errors[-1]) <= \
                    tol_rel_improv * tol_rel_iters * emax:
                        break
            if tol_iters_after_best is not None:
                if iterbest + tol_iters_after_best < it:
                    break
            if maxtime and curtime >= maxtime:
                break
        if callback is not None:
            callback(it, errors, A.T, B)

    if nfixed_a and not np.isscalar(fixed_components_A):
        # Could remove the fixed components from A here, but it's not clear
        # if that's more convenient when anyway they're in B too.
        pass

    if return_time:
        return Abest.T, Bbest, errors, times
    return Abest.T, Bbest, errors


def mcr_als_freeze(sp, initial_A, *, maxiters=100,
                   tol_rel_improv=None, callback=None,
                   freeze_components_B, freeze_scheme,
                   add_residual_components=None, print_progress=True,
                   **kwargs):
    """
    Perform MCR-ALS nonnegative matrix decomposition on the matrix sp.
    The input array is decomposed as A@B where A and B are spectra and
    concentrations or vice versa.

    This version adds gradually applied constraints on the B matrix: The
    values (e.g. concentrations) are iteratively constrained to their mean
    values across all or a subset of the samples. This is useful for unmixing
    of sets of samples (e.g. time courses or from different individuals).
    Parameters are passed throught to mcr_als.

    Parameters
    ----------
    sp : array(nsamples, nfeatures)
        Spectra to be decomposed.
    initial_A : array(ncomponents, nfeatures)
        Initial spectra (or concentrations).
    maxiters : int, optional
        Maximum number of iterations.
    tol_rel_improv : float, optional
        Stop at average relative improvement of less than tol_rel_improv.
    callback : func(it : int, err : list(float), A : array, B : array), optional
        Progress callback for every iteration.
    freeze_components_B : int or [int] or {int: [range, ...]}
        Components that are to be made constant across samples by means of
        growing Lagrange multipliers.
        If int, take that many of the first components. The set/dict forms
        specify component indexes and optionally sample ranges. With the
        latter, an endmember can be frozen to different levels independently
        in different sets of samples.
    freeze_scheme : ndarray(iter) or ndarray(iter, index)
        Lagrange multiplier scheme for forcing the constant components from
        free to constant or constrained. If 2D, the index refers to the order
        of the components in freeze_components_B.
        There will always be a first unconstrained iteration. In the last
        iteration, a value of Inf is allowed and recommended for components
        that are to be made constant.
        Recommended starting point: [1, 3, 10, numpy.inf]
    add_residual_components : int, optional
        Add this many new components based on principal components of
        residuals after the initial MCR-ALS step.
        If you want to freeze a newly added residual component, it first
        needs one iteration with lambda=0 or it will be frozen to 0.
    print_progress : bool, optional
        Print some useful numbers for each iteration.

    Returns
    -------
    A : array(ncomponents, nfeatures)
        Spectra (at lowest error)
    B : array(ncomponents, nsamples)
        Concentrations at lowest error
    error : list(float)
        Mean square error at every iteration
    process_time : list(float)
        Time relative start at each iteration
    """

    starttime = time.process_time()
    nsamp, nfeat = sp.shape
    ncomp = initial_A.shape[0]

    if type(freeze_components_B) is dict:
        f_ranges = [(f, frng) for f, frng
                    in freeze_components_B.items()]
    else:
        if np.isscalar(freeze_components_B):
            freeze_components_B = list(range(freeze_components_B))
        f_ranges = [(f, [range(nsamp)]) for f in freeze_components_B]

    niter = len(freeze_scheme)
    iters = np.geomspace(maxiters ** .75, maxiters, num=niter + 1, dtype=int)
    if tol_rel_improv is None:
        tolrel = [None] * (niter + 1)
    else:
        tolrel = np.geomspace(tol_rel_improv ** .5, tol_rel_improv,
                              num=niter + 1)
    kwargs['return_time'] = True

    cbinfo = [0]
    def extended_cb(it, err, A, B):
        if callback is not None:
            callback(cbinfo[0], err, A, B)
        cbinfo[0] += 1

    def freeze_components(A, B, sp_out, *, lamb, franges):
        for fi, (f, rngs) in enumerate(franges):
            l = lamb if np.isscalar(lamb) else lamb[fi]
            for rng in rngs:
                sp_out[rng, fi] = B[f, rng].mean() * l

    if print_progress:
        print(f"Running initial MCR-ALS, size {sp.shape} / {initial_A.shape}")
    A, B, full_error, full_process_time = mcr_als(
        sp, initial_A, maxiters=iters[-1], tol_rel_improv=tolrel[-1],
        callback=extended_cb, starttime=starttime, **kwargs)
    if print_progress:
        print(f"  MCR-ALS in {full_process_time[-1]:.2f} s, "
              f"{len(full_error)} iters, err {full_error[-1]:.5g}")

    if len(freeze_scheme) == 0:
        return A, B, full_error, full_process_time

    if add_residual_components:
        res = sp - B.T @ A
        pca = pca_nipals(res, add_residual_components)
        A = np.vstack((A, pca * A.max() / pca.max(1, keepdims=True)))
        B = np.vstack((B, np.zeros((add_residual_components, nsamp))))
    # Components may have been added
    ncomp = len(A)

    nfreeze = len(f_ranges)
    fsp = np.hstack((np.zeros((nsamp, nfreeze)), sp.copy()))
    fA = np.hstack((np.zeros((ncomp, nfreeze)), A.copy()))

    for i, lamb in enumerate(freeze_scheme):
        # On the last pass, identify (fi, f) that are to be removed from
        # fsp, fA, f_ranges, f_comps, nfreeze and that will be marked
        # as non-mutable.
        if (i == len(freeze_scheme) - 1) and np.any(np.isinf(lamb)):
            nonfrozen = np.ones(ncomp, dtype=bool)
            frozenixs = []
            for fi, (f, rngs) in enumerate(f_ranges):
                l = lamb if np.isscalar(lamb) else lamb[fi]
                if np.isinf(l):
                    nonfrozen[f] = False
                    frozenixs.append(fi)
                    for rng in rngs:
                        B[f, rng] = B[f, rng].mean()
            fsp = np.delete(fsp, frozenixs, axis=1)
            fA = np.delete(fA, frozenixs, axis=1)
            f_ranges = [x for i, x in enumerate(f_ranges)
                        if i not in frozenixs]
            nfreeze = len(f_ranges)
            if not np.isscalar(lamb):
                lamb = np.delete(lamb, frozenixs)
        else:
            nonfrozen = slice(None)

        f_comps = [f for f, rngs in f_ranges]
        fA[f_comps, :nfreeze] = np.eye(nfreeze) * lamb
        freeze_components(None, B, fsp, lamb=lamb, franges=f_ranges)

        if print_progress:
            print(f"Running MCR-ALS, λ={lamb}, size {fsp.shape} / {fA.shape}")
        fA, B, error, process_time = mcr_als(
            fsp, fA, maxiters=iters[i], tol_rel_improv=tolrel[i],
            fixed_features_A=nfreeze,
            modify_before_Astep=partial(
                freeze_components, lamb=lamb, franges=f_ranges),
            mutable_components_B=nonfrozen,
            initial_B=B,
            callback=extended_cb, starttime=starttime, **kwargs)
        full_error.extend(error)
        full_process_time.extend(process_time)
        if print_progress:
            err = ((sp - B.T @ fA[:, nfreeze:])**2).sum()
            print(f"  time={process_time[-1]:.2f} s, "
                  f"{len(error)} iters, err {error[-1]:.5g} (tot {err:.5g})")

    A = fA[:, nfreeze:]
    return A, B, full_error, full_process_time




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 14:01:08 2020

@author: carl
"""

import os.path
import numpy as np
import scipy.signal, scipy.io
import pandas as pd
from pymatreader import read_mat
import h5py

from .opusreader import OpusReader
from .ptirreader import PtirReader
from .omnicreader import OmnicReader
from ..algorithms.util import pixels_fit
from .. import octavvs_version

class SpectralData:
    """
    The list of current files and the raw data loaded from a hyperspectral
    image file in the GUIs.
    """
    def __init__(self):
        self.foldername = '' # Root dir
        self.filenames = []  # With full paths
        self.curFile = ''    # The currently loaded file
        self.wavenumber = None # array in order low->high
        self.wmin = 800
        self.wmax = 4000
        self.raw = np.empty((0,0)) # data in order (pixel, wavenumber)
        self.wh = (0, 0)  # Width and height in pixels
        self.pixelxy = None # Positions of pixels in image
        self.images = None # list of Image loaded from raw data file
        self.filetype = None # str from [txt, mat, opus, ptir]
        self.ptir_h5info = None # Possibly populated when loading PTIR file

    def set_width(self, w):
        try:
            w = int(w)
        except ValueError:
            w = 0
        if w <= 0:
            return False
        h = (self.raw.shape[0] + w - 1) // w
        # if w * h != self.raw.shape[0]:
        #     return False
        self.wh = (w, h)
        return True

    def set_height(self, h):
        if self.set_width(h):
            self.wh = (self.wh[1], self.wh[0])
            return True
        return False

    def get_xy(self):
        """
        Get pixel coordinates, regardless of data format.

        Returns
        -------
        Coordinates : numpy.ndarray
            Array of size (npixels, 2) with x, y pairs.
        """
        if self.pixelxy is not None:
            return self.pixelxy
        elif self.wh is not None and self.wh[0] and self.wh[1]:
            return np.indices((self.wh[1], self.wh[0])).reshape(
                (2, -1)).T[:len(self.raw),::-1]
        else:
            return np.vstack(range(len(self.raw)), [0] * len(self.raw)).T

    def extract_data_from_matrix(self, mx, meta):
        """
        Extract wavenumbers, raw data and image dimensions from some kind
        of data matrix, with or without additional metadata

        Parameters
        ----------
        mx : numpy.ndarray or None
            raw matrix data, possibly with wavenumbers and annotations.
        meta : dict
            additional metadata, possibly including mx itself

        Returns
        -------
        raw : numpy.ndarray
            matrix of floats
        wn : numpy.ndarray
            sorted array of wavenumbers
        wh : tuple or None
            image dimensions (width and height)
        """
        if meta is None:
            meta = {}
        wh = None

        # Check for some particular format(s)
        if 'ImR' in meta and 'lambda' in meta:
            # Matlab file with data in ImR, wavelengths in lambda
            # (encountered with hyperspectral data in Lund)
            mx = meta['ImR']
            if mx.ndim == 2:
                raw = mx
            else:
                assert mx.ndim == 3
                wh = [mx.shape[1], mx.shape[0]]
                raw = mx.reshape((-1, mx.shape[2]))
            wn = meta['lambda'].flatten()
            assert len(wn) == raw.shape[1]
            return raw, wn, wh
        if 'wavenumber' in meta and 'y' in meta:
            # Quasar.mat - and we ought to load map_x/y and figure out a
            # wh rectangle or fall back to pixelxy if necessary.
            wn = meta['wavenumber'].flatten()
            raw = meta['y']
            ...
            return raw, wn, None

        if mx is None:
            # Assume data are in the biggest matrix in the file
            mx = max(meta.items(), key=lambda k: np.size(k[1]))[1]

        if mx.ndim == 3 and (np.array(mx.shape) > 1).all():
            wh = (mx.shape[0], mx.shape[1])
            mx = mx.reshape((-1, mx.shape[2]))
        if mx.ndim != 2 or (np.array(mx.shape) < 2).any():
            raise RuntimeError(
                'file does not appear to describe an FTIR image matrix')

        if 'wh' in meta:
            wh = meta['wh'].flatten()

        # Assume wavenumbers are a) first row or b) first column
        if mx.dtype == object:
            # Find columns with only numbers. Removed strings ought to
            # be saved as annotations.
            alltypes = np.vectorize(lambda x: np.issubdtype(
                type(x), np.number))(mx)
            numcols = [np.all(alltypes, axis=0), np.all(alltypes, axis=1)]
            wns = [wn.astype(float, casting='unsafe') for wn in
                   [mx[0, numcols[0]], mx[numcols[1], 0]]]
        else:
            wns = [mx[0, :], mx[:, 0]]
        # Find the direction with most sorted wavenumbers
        for i, wn in enumerate(wns):
            deltas = np.diff(wn)
            if np.any(deltas <= 0) and np.any(deltas >= 0):
                wns[i] = []
        if len(wns[0]) < 3 and len(wns[1]) < 3:
            raise RuntimeError(
                'First column or row must contain sorted wavenumbers')
        # Transpose if needed
        if len(wns[1]) > len(wns[0]):
            mx = mx.T
            mxt = 1
        else:
            mxt = 0
        wn = wns[mxt]
        # Remove the non-numeric stuff
        if mx.dtype == object:
            nonnum = mx[0, ~numcols[mxt]]
            if len(nonnum):
                print(f'Removing non-numeric fields: {nonnum}')
            raw = mx[1:, numcols[mxt]].astype(float)
        else:
            raw = mx[1:]

        if raw.dtype != np.float32 and raw.dtype != np.float64:
            print(f'Loaded as {raw.dtype}, converting to float64')
            wn = wn.astype(float)
            raw = raw.astype(float)
        return raw, wn, wh

    def read_matrix(self, filename):
        """
        Read data from a file, with some error checking. The object is modified
        only if the file is successfully loaded.
        """
        wh = None
        xy = None
        images = None
        raw = None
        ptir_h5info = None
        fext = os.path.splitext(filename)[1].lower()
        if fext == '.mat':
            filetype = 'mat'
            try:
                s = read_mat(filename)
            except TypeError:
                # Workaround for uint16_codec bug (pymatreader
                # assumes mio5, not mio4)
                s = scipy.io.loadmat(filename)
            raw, wn, wh = self.extract_data_from_matrix(None, s)
        elif fext in ['.txt', '.csv', '.xls', '.xlsx']:
            if fext in ['.xls', '.xlsx']:
                filetype = 'xls'
                ss = pd.read_excel(filename, header=None).to_numpy()
            else:
                filetype = 'txt'
                ss = pd.read_csv(filename, sep=None, engine='python',
                                 header=None, comment="#").to_numpy()
            raw, wn, wh = self.extract_data_from_matrix(ss, None)
        else:
            if fext == '.ptir' or fext == '.hdf':
                filetype = 'ptir'
                reader = PtirReader()
                reader.load(filename, keep_copyable=True)
                xy = reader.xy
                ptir_h5info = reader.h5info
            elif fext in ['.spa', '.spg']:
                filetype = 'omnic'
                reader = OmnicReader(filename)
            else:
                filetype = 'opus'
                reader = OpusReader(filename)
            raw = reader.AB
            wn = reader.wavenum
            wh = reader.wh
            images = reader.images

        # Ensure ascending wavenumbers
        assert len(wn) == raw.shape[1]
        if wn[0] > wn[1]:
            wn = wn[::-1]
            raw = raw[:, ::-1]
        diffsign = np.sign(np.diff(wn))
        if (diffsign >= 0).any() and (diffsign <= 0).any():
            raise RuntimeError('wavenumbers must be sorted')

        # Guess the image shape if square
        npix = raw.shape[0]
        if wh is not None:
            # print(f'image size {wh}')
            if len(wh) != 2:
                raise RuntimeError('Image size in "wh" must have length 2')
            wh = (int(wh[0]), int(wh[1]))
            if not pixels_fit(npix, wh):
                raise RuntimeError('Image size in "wh" does not match data size')
        elif xy is not None:
            # We should do some clever deduction of size and add code
            # to deduce the best rectangular grid from xy coordinates,
            # but for now we keep it simple.
            yy = np.diff(xy[:, 0])
            yy = yy - yy.mean()
            yy = yy / np.median(yy)
            guessh = (yy < .25 * yy.min()).sum() + 1
            if guessh > npix**.2 and guessh < npix **.8:
                wh = ((npix + guessh - 1) // guessh, guessh)
                print(f'assuming image size {wh} from xy data')
        if wh is None:
            if npix == self.wh[0] * self.wh[1]:
                wh = self.wh
            else:
                res = int(np.sqrt(npix))
                if npix == res * res:
                    wh = (res, res)
                    print(f'assuming square image {wh}')
                else:
                    wh = (npix, 1)
                    print(f'assuming non-grid data {wh}')
        self.wh = wh

        self.raw = raw
        self.wavenumber = wn
        self.wmin = self.wavenumber.min()
        self.wmax = self.wavenumber.max()
        self.images = images
        self.pixelxy = xy
        self.curFile = filename
        self.filetype = filetype
        self.ptir_h5info = ptir_h5info

    def save_matrix_matlab_ab(self, filename, wn, ydata, metadata):
        ab = np.hstack((wn[:, None], ydata.T))
        abdata = {'AB': ab, 'wh': self.wh }
        if self.pixelxy is not None:
            abdata['xy'] = self.pixelxy
        for k, v in metadata.items():
            abdata[k] = v
        scipy.io.savemat(filename, abdata)

    def save_matrix_matlab_quasar(self, filename, wn, ydata, metadata):
        out = {'y': ydata, 'wavenumber': wn}
        if self.pixelxy is not None:
            map_x, map_y = np.array(self.pixelxy).T
        else:
            map_x = np.tile(range(self.wh[1]), self.wh[0])
            map_y = np.repeat(range(self.wh[0]), self.wh[1])
        out['map_x'] = map_x[:, None]
        out['map_y'] = map_y[:, None]
        for k, v in metadata.items():
            out[k] = v
        scipy.io.savemat(filename, out)

    def save_matrix_pandas(self, filename, wn, ydata, fmt, metadata):
        df = pd.DataFrame(ydata.T, index=wn)
        kpad = "#" if fmt == "csv" else ""
        mdf = pd.DataFrame(metadata.values(),
                           index=[kpad + k for k in metadata])
        df = pd.concat([mdf, df])
        if fmt == 'csv':
            df.to_csv(filename, header=False, float_format="%.16g")
        elif fmt == 'xlsx':
            df.to_excel(filename)

    def save_matrix_ptir(self, filename, wn, ydata, metadata=None):
        f = h5py.File(filename, mode='w')

        f.attrs['DocType'] = b'IR'
        f.attrs['SoftwareVersion'] = octavvs_version.encode('utf8')

        digits = int(np.log10(len(ydata))) + 1
        for i, d in enumerate(ydata):
            mm = f.create_group(f'Measurement{i+1:0{digits}}')
            # if self.pixelxy is not None:
            #     x, y = self.pixelxy[i]
            # else:
            #     x = i % self.wh[0]
            #     y = i // self.wh[0]
            x = metadata["X"][i]
            y = metadata["Y"][i]
            mm.attrs['LocationX'] = [x]
            mm.attrs['LocationY'] = [y]
            if "Filename" in metadata:
                mm.attrs["Filename"] = [metadata["Filename"][i]]

            if not i:
                specval = mm.create_dataset('Spectroscopic_Values',
                                            data=wn[None, :])

            ch = mm.create_group('Channel_000')
            rd = ch.create_dataset('Raw_Data', data=d[None, :])
            rd.attrs['Spectroscopic_Values'] = specval.ref

        if self.images is not None and len(self.images):
            imgs = f.create_group('Images')
            digits = max(3, int(np.log10(len(self.images))) + 1)
            for i, im in enumerate(self.images):
                imds = imgs.create_dataset(f'Image_{i:0{digits}}',
                                           data=im.data[::-1])
                imds.attrs['Label'] = np.bytes_(im.name)
                imds.attrs['PositionX'] = [im.xy[0]]
                imds.attrs['PositionY'] = [im.xy[1]]
                imds.attrs['SizeWidth'] = [im.wh[0]]
                imds.attrs['SizeHeight'] = [im.wh[1]]

        f.close()

    def save_matrix(self, filename, fmt, wn=None, ydata=None,
                    metadata=None):
        """
        Save processed data in some format.

        Parameters
        ----------
        filename : str
            Name of the output file.
        fmt : str, optional
            File format: 'ab', 'quasar', 'ptirlite', 'ptir', 'csv', 'xlsx'.
            The default is 'ab', a MATLAB file with array 'AB'
            with wavenumbers in the first _column_,
            image size in 'wh' and optionally pixel coordinates in 'xy'.
            The 'quasar' format has wavenumbers in a separate array.
            The 'ptirlite' format is a reduced variant of the HDF5 format used
            by PTIR Studio.
            The 'ptir' format requires that the loaded file was PTIR.
        wn : 1D array, optional
            Wavenumber vector. The default is self.wn.
        ydata : 2D array, optional
            Data in (pixel, wn) order. The default is self.raw.
        metadata : dict, optional
            Data to be included in the output, e.g. as comment lines before
            the data.

        Returns
        -------
        None.

        """
        if wn is None:
            wn = self.wn
        if ydata is None:
            ydata = self.raw
        if metadata is None:
            pxy = self.get_xy().T
            metadata = {"X": pxy[0], "Y": pxy[1]}
        if fmt == 'ab':
            self.save_matrix_matlab_ab(filename, wn, ydata, metadata=metadata)
        elif fmt == 'quasar':
            self.save_matrix_matlab_quasar(
                filename, wn, ydata, metadata=metadata)
        elif fmt == 'ptirlite':
            self.save_matrix_ptir(filename, wn, ydata, metadata=metadata)
        elif fmt == 'ptir':
            if self.ptir_h5info is None:
                print("Falling back to PTIR-lite format")
                self.save_matrix_ptir(filename, wn, ydata, metadata=metadata)
            else:
                PtirReader.copy_and_save(filename, ydata, self.ptir_h5info)
        elif fmt == 'csv' or fmt == 'xlsx':
            self.save_matrix_pandas(
                filename, wn, ydata, fmt, metadata=metadata)
        else:
            raise ValueError('Unknown save file format')





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  4 16:24:50 2020

@author: carl
"""

import os.path
import importlib
import numpy as np
from scipy.interpolate import PchipInterpolator
#from scipy.io import loadmat
from pymatreader import read_mat
import pandas as pd


def load_reference(wn, what=None, filename=None):
    """
    Load and normalize a spectrum from a Matlab file, interpolating at
    the given points. The reference is assumed to cover the entire range
    of wavenumbers.

    Parameters
    ----------
    wn : ndarray
        wavenumbers at which to get the spectrum.
    what : string, optional
        Type of reference to get, corresponding to a file in the
        'reference' directory.
    filename : string, optional
        The name of an arbitrary Matlab/CSV file to load data from.
        For .mat, the data must be in a matrix called AB, with wavenumbers
        in the first column.
        For CSV, the wavenumbers must be in the first row and the spectrum
        in the second row.

    Returns
    -------
    Spectrum at the points given by wn.

    """
    if (what is None) == (filename is None):
        raise ValueError("Either 'what' or 'filename' must be specified")
    if what is not None:
        filename = importlib.resources.files("octavvs").joinpath(
            "reference_spectra", what + ".mat")
    if os.path.splitext(filename)[1].lower() == ".mat":
        ref = read_mat(filename)['AB']
    else:
        ref = pd.read_csv(filename, sep=None, comment='#', engine="python",
                          header=None)
        ref = ref.to_numpy()
        if ref.shape[0] < ref.shape[1]:
            ref = ref.T
        if ref.shape[1] != 2:
            raise RuntimeError(
                "CSV reference error: wavenumbers must be in the"
                " first row/col and the spectrum in the second row/col."
                f" (got: {ref.shape})")
    # Handle the case of high-to-low since the interpolator requires low-to-high
    d = 1 if ref[0,0] < ref[-1,0] else -1
    ref = PchipInterpolator(ref[::d,0], ref[::d,1])(wn)
    return ref #/ ref.max()

def nonnegative(y, fracspectra=0, fracvalues=0):
    """
    Make a matrix of spectral data nonnegative by shifting all the spectra
    up by the same computed amount, followed by setting negative values to 0.
    The shift is chosen such that at most fracspectra of the spectra get more
    than fracvalues of their intensities set to zero.
    Parameters
    ----------
    y : ndarray(pixel, wavenumber)
        Intensities.
    fracspectra : float
        Unheeded fraction of the spectra.
    fracvalues : float
        Maximal fraction of points to clip at 0.
    Returns
    -------
    out : ndarray(pixel, wavenumber)
        Shifted spectra in the same format as y.
    """
    s = int(fracspectra * y.shape[0])
    v = int(fracvalues * y.shape[1])
    if s == 0 or v == 0:
        return y - np.min(y.min(), 0)
    if s >= y.shape[0] or v >= y.shape[1]:
        return np.maximum(y, 0)
    yp = np.partition(y, v, axis=1)[:,v]
    a = np.partition(yp, s)[s]
    return np.maximum(y - a if a < 0 else y, 0)

def pca_nipals(x, ncomp, tol=1e-5, max_iter=1000, copy=True,
               explainedvariance=None):
    """
    NIPALS algorithm for PCA, based on the code in statmodels.multivariate
    but with optimizations as in Bassan's Matlab implementation to
    sacrifice some accuracy for speed.

    Parameters
    ----------
    x : ndarray(samples, features)
        Data.
    ncomp : int
        Number of PCA components to return.
    tol : float
        Relative tolerance.
    copy : bool
        If false, destroy the input matrix x.
    explainedvariance : None or float
        Stop after this fraction of the total variance is explained.

    Returns
    ----------
    vecs : int
        PCA loadings as rows
    """
    if copy:
        x = x.copy()
    if explainedvariance is not None and explainedvariance > 0:
        varlim = (x * x).sum() * (1. - explainedvariance)
    else:
        varlim = 0
    npts, nvar = x.shape
    vecs = np.empty((ncomp, nvar))
    for i in range(ncomp):
        factor = np.ones(npts)
        for j in range(max_iter):
            vec = x.T @ factor #/ (factor @ factor)
            vec = vec / np.sqrt(vec @ vec)
            f_old = factor
            factor = x @ vec #/ (vec @ vec)
            f_old = factor - f_old
            if tol > np.sqrt(f_old @ f_old) / np.sqrt(factor @ factor):
                break
        vecs[i, :] = vec
        if i < ncomp - 1:
            x -= factor[:,None] @ vec[None,:]
            if varlim > 0 and (x * x).sum() < varlim:
                return vecs[:i+1, :]
    return vecs

def find_wn_ranges(wn, ranges):
    """
    Find indexes corresponding to the beginning and end of a list of ranges
    of wavenumbers. The wavenumbers have to be sorted in either direction.

    Parameters
    ----------
    wn : ndarray
        Wavenumbers
    ranges : ndarray(n, 2)
        Desired wavenumber ranges in order [low, high]
    Returns
    -------
    indexes : ndarray(n, 2)
        Indexes of wavenumbers delimiting the input ranges
    """
    if isinstance(ranges, list):
        ranges = np.array(ranges)
    if(wn[0] < wn[-1]):
        return np.stack((np.searchsorted(wn, ranges[:,0]),
                         np.searchsorted(wn, ranges[:,1], 'right')), 1)
    return len(wn) - np.stack((np.searchsorted(wn[::-1], ranges[:,1], 'right'),
                              np.searchsorted(wn[::-1], ranges[:,0])), 1)

def pixels_fit(npixels, wh):
    "Check that n pixels fit snugly within w * h"
    return wh[0] * (wh[1] - 1) < npixels <= wh[0] * wh[1]


def subtract_background(y, background):
    """
    Subtract such an amount of a background spectrum that the resulting
    spectra become maximally smooth.

    Parameters
    ----------
    y : ndarray
        1D or 2D array of spectra.
    background : ndarray
        1D reference spectrum of the same length as y.

    Returns
    -------
    ret : ndarray
        Input spectra minus background.
    """

    dh = background[:-1] - background[1:]
    dy = y[:,:-1] - y[:,1:]
    dh2 = np.cumsum(dh * dh)
    dhdy = np.cumsum(dy * dh, 1)

    az = dhdy[:,-1] / dh2[-1]
    ret = y - az[:, None] @ background[None, ...]
    return ret
